/** -------------------------------------------------------------------
  *      maintainer: Marcus Wolschon <Suran@gmx.net>
  * -------------------------------------------------------------------
  * projects:
  *	- Jaugment II
  *       part: Swing-UI
  *	   used in release: current      used in version: current
  * -------------------------------------------------------------------
  * Copyright:   Copyright (c) 2000 by Marcus Wolschon <Suran@gmx.net>
  * -------------------------------------------------------------------
  * changes:
  * 0.0.1 (2000-12-09, Marcus Wolschon<Suran@gmx.net>)
  *       - created
  * 0.0.2 (2000-12-10, Marcus Wolschon<Suran@gmx.net>)
  *       - is a JINI-service
  *       - deploys all services it finds as SOAP-services
  *       - deploys itself
  *       - Apache-SOAP-Server is configurable
  *       - tries to deploy the admin-objects too
  * -------------------------------------------------------------------
  * TODO:
  *       - be Administratable (type-mamings, map all/some named service, undeply services, wrap recursive, use the beans for Entries+optionally wrap that beans...)
  *       - map a getAttributes-method and the administrative object
  *       - write, compile and register wrappers for all SOAP-Services found
  */
package jaugment2.soap2jini;

  //JDK
import java.rmi.RemoteException;
import java.rmi.Remote;
import java.rmi.MarshalledObject;
import java.util.StringTokenizer;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Date;
import java.util.Enumeration;
import java.net.URL;
import java.net.InetAddress;
import java.lang.reflect.*;
import java.io.PrintWriter;
  //JINI
import net.jini.core.lookup.ServiceItem;
import net.jini.core.lookup.ServiceID;
import net.jini.core.lease.LeaseDeniedException;
import net.jini.core.lease.UnknownLeaseException;
import net.jini.core.lease.Lease;
import net.jini.core.event.RemoteEvent;
import net.jini.core.event.RemoteEventListener;
import net.jini.core.event.UnknownEventException;
import net.jini.admin.Administrable;
import net.jini.core.entry.Entry;
import net.jini.core.discovery.LookupLocator;
import net.jini.discovery.DiscoveryListener;
import net.jini.discovery.DiscoveryEvent;
  //Apache/IBM-Soap
import org.apache.soap.*;
import org.apache.soap.server.DeploymentDescriptor;
import org.apache.soap.server.ServiceManagerClient;
  //JAugment
import jaugment2.jini.simpleService.simpleService;
import jaugment2.jini.SimpleLandlord;
import jaugment2.jini.SimpleLandlord.LeaseListener;
import jaugment2.jini.SimpleLandlord.EventRegistration;
import jaugment2.helpers.Logger;
import jaugment2.jini.LookupEventListener;
import jaugment2.jini.LookupEventEvent;

//helper-mark for configure to determine that rmic needs to be executed for this file
//rmimakehelper

public class Soap2JINI extends simpleService implements Remote, DiscoveryListener, LookupEventListener
{

  //--------------------- local lookup ------------------------------------------------------//

//TODO: RemoteEventListener to listen for events from our registrars


  public void ServiceFoundEvent(LookupEventEvent e)
  {
   Logger.debug.println("DEBUG: [Soap2JINI] ServiceFoundEvent");
   if(smc != null)
   if(e.item != null)
   if(e.item.service != null)
   if(convertedJINIServices.get(e.item.serviceID) == null)
     {

      convert(e.item);
     }

  }

  public void ServiceGoneEvent(LookupEventEvent e)
  {
   if(smc != null)
   if(convertedJINIServices.get(e.item.serviceID) != null)
     Logger.debug.println("DEBUG: [Soap2JINI] ServiceGoneEvent");
  }

  public void ServiceChangedEvent(LookupEventEvent e)
  {
   if(smc != null)
   if(convertedJINIServices.get(e.item.serviceID) != null)
     Logger.debug.println("DEBUG: [Soap2JINI] ServiceChangedEvent");
  }


  //--------------------- initiaslisation&administration ------------------------------------------------------//

  private Soap2JINIProxy myproxy =null;


  /**
   * Load our state (e.g. attrubutes, service-id) from persistent storage
   */
  private void loadMe()
  {
      String filename = System.getProperty("jaugment2.soap2jini.Soap2JINI.persistentStateFilename");
      if(filename==null)
        {
         filename = "Soap2JINIState.dat";
         System.setProperty("jaugment2.soap2jini.Soap2JINI.persistentStateFilename", filename);
        }

      persistentState.setProperty("persistentStateFilename", filename);

   load();


   if(persistentState.get("attr")==null)
     {
      Entry[] attrSets = new Entry[]
                   {
                     new net.jini.lookup.entry.Name("Soap2JINI"),
                     new net.jini.lookup.entry.Comment("a bridge between SOAP and JINI"),
                     new net.jini.lookup.entry.ServiceInfo("Soap2JINI", "Marcus Wolschon <Suran@gmx.net>", "Marcus Wolschon <Suran@gmx.net>", "0.0.1", "standart", "00000"),

//no representation, at least yet                     DirectoryServiceRepresentation.getUIDescriptor()
                    };
      persistentState.put("attr", attrSets);
     }

   save();

  }


  public Soap2JINI() throws RemoteException
  {
   this(null);
  }

  public Soap2JINI(LookupLocator locallocator) throws RemoteException
  {
   Logger.debug.println("DEBUG: [Soap2JINI] creating myself...");
   try
     {

//      landLord = new SimpleLandlord(simpleLeaselistener);

      init();

      persistentState.setProperty("groups", "public, Suran");
      persistentState.setProperty("locators", "jini://localhost, jini://127.0.0.1");

      myproxy = new Soap2JINIProxy(this);

      loadMe();

      Entry[] attrSets = (Entry[])persistentState.get("attr");

      this.init(myproxy, attrSets, this, locallocator);

     }
   catch(Exception e)
     {
      Logger.err.println("ERROR: [Soap2JINI] error in constructor because of : "+e.getMessage());
      e.printStackTrace(Logger.err);
      this.smc = null;
      try
        {
         stop(false);
        }
      catch(Exception ex)
        {
         ex.printStackTrace(Logger.err);
        }
     }
  }


    /**
     * Called when one or more lookup service registrars has been discovered.
     * The method should return quickly; e.g., it should not make remote
     * calls.
     *
     * @param e the event that describes the discovered registrars
     */
    public void discovered(DiscoveryEvent e)
    {
     Logger.debug.println("DEBUG: [Soap2JINI] found "+e.getRegistrars().length+" registrars");
    }

    /**
     * Called when one or more lookup service registrars has been discarded.
     * The method should return quickly; e.g., it should not make remote
     * calls.
     *
     * @param e the event that describes the discarded registrars
     */
    public void discarded(DiscoveryEvent e)
    {
     Logger.debug.println("DEBUG: [Soap2JINI] discarded "+e.getRegistrars().length+" registrars");
    }

 /** restart this service once it has been stopped
   * @see stop
   */
 public void restart() throws Exception
 {
  //TODO: reregister, start engine
 }

 /** stop this service. If exit is false all references to this
   * service are still valid but it has deregistered everywhere.
   * (this includes the RMI-activation-daemon)
   * @see restart
   * @see destroy
   * @param exit stop running
   */
 public void stop(boolean exit) throws Exception
 {
  deinit();
  //TODO: deregister, stop engine(, exit this thread)
 }

 /** stop this service, destroy any persistent storage and exit
   */
 public void destroy() throws RemoteException
 {
  deinit();
  //TODO: deregister, stop engine, delete persistent storage, exit this thread
 }


 /*public void main(String argc[])
 {
 }*/

  //--------------------- functionality ------------------------------------------------------//

 ServiceManagerClient smc = null;
 private void init()
 {
  try
    {
     URL server = null;
     try
       {
        server = new URL(System.getProperty("jaugment2.soap2jini.Soap2JINI.serverurl"));
       }
     catch(Exception e)
       {
       }
     if(server == null)
       server = new URL("http://localhost/soap/");

     smc = new ServiceManagerClient(server);
     String others[] = smc.list();
     //TODO: convert them if needed to JINI-services
    }
  catch(Exception e)
    {
     Logger.err.println("ERROR: [Soap2JINI] could not initialize because of : "+e.getMessage());
//     e.printStackTrace(Logger.err);
     smc = null;
    }
 };

 private Hashtable convertedJINIServices = new Hashtable();
 private Hashtable convertedJINIServicesAdmins = new Hashtable();


 private void convert( ServiceItem item)
 {
  try
    {
     if(smc == null)
      return;

     if(item.service == null)
       return;

     if(convertedJINIServices.get(item.serviceID) != null)
       {// service has changes: remove it first
        DeploymentDescriptor desc = (DeploymentDescriptor)convertedJINIServices.get(item.serviceID);
        smc.undeploy(desc.getID());
       };
     Class proxyclass = item.service.getClass();
     if(proxyclass == null)
       return;

     Object p2 = convert2(proxyclass);
     if(p2 != null)
       convertedJINIServices.put(item.serviceID, p2);

     if(item.service instanceof Administrable)
      {
       if(convertedJINIServicesAdmins.get(item.serviceID) != null)
         {// service has changes: remove it first
          DeploymentDescriptor desc = (DeploymentDescriptor)convertedJINIServicesAdmins.get(item.serviceID);
          smc.undeploy(desc.getID());
         };
       Class adminclass = ((Administrable)item.service).getAdmin().getClass();
       if(adminclass.getConstructor(new Class[0]) != null)
          {
           Object p = convert2(adminclass);
           if(p != null)
             convertedJINIServices.put(item.serviceID, p);
          }
      }

    }
  catch(Exception e)
    {
     e.printStackTrace(Logger.err);
    }

 }

 private DeploymentDescriptor convert2( Class proxyclass)
 {

  try
    {
     Method proxymethods[] = proxyclass.getDeclaredMethods();

     DeploymentDescriptor desc = new DeploymentDescriptor();
     desc.setID(getClass().getName()+"-"+proxyclass.getName());
     desc.setIsStatic(false);
     //TODO: setMappings


     String methods[] = new String[proxymethods.length];
     for(int i=0; i<methods.length; i++)
       methods[i] = proxymethods[i].getName();

     desc.setMethods(methods);

     desc.setProviderClass(proxyclass.getName());
     desc.setProviderType(DeploymentDescriptor.PROVIDER_JAVA);
     desc.setScope(DeploymentDescriptor.SCOPE_SESSION);
     desc.toXML(new PrintWriter(Logger.debug));


     smc.deploy(desc);
     return(desc);
    }
  catch(Exception e)
    {
     e.printStackTrace(Logger.err);
    }
  return(null);
 }

}